from unittest.mock import patch

import pytest

from ggshield.verticals.hmsl.secret_manager.hashicorp_vault.api_client import (
    VaultAPIClient,
)
from ggshield.verticals.hmsl.secret_manager.hashicorp_vault.exceptions import (
    VaultForbiddenItemError,
    VaultNotFoundItemError,
    VaultPathIsNotADirectoryError,
)
from ggshield.verticals.hmsl.secret_manager.hashicorp_vault.models import (
    VaultKvMount,
    VaultSecrets,
)
from tests.unit.conftest import my_vcr


@pytest.fixture(scope="module")
def vault_api_client():
    """
    Get a Vault API client with all rights.
    """
    return VaultAPIClient(
        vault_url="http://localhost:8200",
        api_token="my_vault_token",  # as defined in scripts/hmsl/hashicorp_vault.py
    )


@pytest.fixture(scope="module")
def restricted_vault_api_client():
    """
    Get a Vault API client with restricted permissions
    """
    return VaultAPIClient(
        vault_url="http://localhost:8200",
        api_token="restricted_token",  # as defined in scripts/hmsl/hashicorp_vault.py
    )


@my_vcr.use_cassette(
    "test_hmsl_secret_manager_hashicorp_vault_get_kv_mounts_api_endpoint.yaml",
    ignore_localhost=False,
)
def test_get_kv_mounts_api_endpoint(vault_api_client):
    """
    GIVEN a Vault instance and an API client
    WHEN calling get_kv_mounts
    THEN I get all the kv mounts of this instance
    """
    kv_mounts = list(vault_api_client.get_kv_mounts())

    assert kv_mounts == [
        VaultKvMount(name="secret", version="2"),
        VaultKvMount(name="secret_v1", version="1"),
    ]


@pytest.mark.parametrize(
    "vault_path, expected_results",
    [
        ("", ["b2b", "b2c"]),
        ("b2b", ["web_app", "worker"]),
        ("b2b/worker", ["config.env"]),
        ("b2b/web_app", ["config.env", "prod"]),
        ("b2c", ["web_app", "worker"]),
        ("b2c/worker", ["config.env"]),
        ("b2c/web_app", ["config.env", "prod"]),
        ("/b2b", ["web_app", "worker"]),
        ("/b2b/worker", ["config.env"]),
        ("/b2b/web_app", ["config.env", "prod"]),
        ("/b2c", ["web_app", "worker"]),
        ("/b2c/worker", ["config.env"]),
        ("/b2c/web_app", ["config.env", "prod"]),
    ],
)
def test_list_kv_items_v1(vault_api_client, vault_path, expected_results):
    """
    GIVEN a Vault instance and an API client
    WHEN calling list_kv_items on a v1 mount
    THEN I get all the items
    """
    kv_mount = VaultKvMount(name="secret_v1", version="1")

    with my_vcr.use_cassette(
        f"test_hmsl_secret_manager_hashicorp_vault_list_kv_items_v1_{vault_path.replace('/', '_')}.yaml",
        ignore_localhost=False,
    ):
        kv_items = list(vault_api_client.list_kv_items(kv_mount, vault_path))

    assert kv_items == expected_results


@my_vcr.use_cassette(
    "test_hmsl_secret_manager_hashicorp_vault_list_kv_items_not_a_directory.yaml",
    ignore_localhost=False,
)
def test_list_kv_items_not_a_directory(vault_api_client):
    """
    GIVEN a Vault instance and an API client
    WHEN calling list_kv_items with a path that is actually a file
    THEN VaultPathIsNotADirectoryError is raised
    """
    kv_mount = VaultKvMount(name="secret", version="2")

    with pytest.raises(VaultPathIsNotADirectoryError):
        vault_api_client.list_kv_items(kv_mount, "b2c/worker/config.env")


@my_vcr.use_cassette(
    "test_hmsl_secret_manager_hashicorp_vault_list_kv_items_forbidden.yaml",
    ignore_localhost=False,
)
def test_list_kv_items_forbidden(restricted_vault_api_client):
    """
    GIVEN a Vault instance and an API client
    WHEN calling list_kv_items on a path you don't have access to
    THEN VaultForbiddenItemError is raised
    """
    kv_mount = VaultKvMount(name="secret", version="2")

    with pytest.raises(VaultForbiddenItemError):
        restricted_vault_api_client.list_kv_items(kv_mount, "")


@pytest.mark.parametrize(
    "vault_path, expected_results",
    [
        ("", ["b2b", "b2c"]),
        ("b2b", ["web_app", "worker"]),
        ("b2b/worker", ["config.env"]),
        ("b2b/web_app", ["config.env", "prod"]),
        ("b2c", ["web_app", "worker"]),
        ("b2c/worker", ["config.env"]),
        ("b2c/web_app", ["config.env", "prod"]),
        ("/b2b", ["web_app", "worker"]),
        ("/b2b/worker", ["config.env"]),
        ("/b2b/web_app", ["config.env", "prod"]),
        ("/b2c", ["web_app", "worker"]),
        ("/b2c/worker", ["config.env"]),
        ("/b2c/web_app", ["config.env", "prod"]),
    ],
)
def test_list_kv_items_v2(vault_api_client, vault_path, expected_results):
    """
    GIVEN a Vault instance and an API client
    WHEN calling list_kv_items on a v2 mount
    THEN I get all the items
    """
    kv_mount = VaultKvMount(name="secret", version="2")

    with my_vcr.use_cassette(
        f"test_hmsl_secret_manager_hashicorp_vault_list_kv_items_v2_{vault_path.replace('/', '_')}.yaml",
        ignore_localhost=False,
    ):
        kv_items = list(vault_api_client.list_kv_items(kv_mount, vault_path))

    assert kv_items == expected_results


@pytest.mark.parametrize(
    "vault_path, expected_results",
    [
        (
            "b2b/web_app/prod/config.env",
            [("b2b/web_app/prod/config.env/PROD_STUFF", "test")],
        ),
        (
            "b2b/worker/config.env",
            [
                ("b2b/worker/config.env/ANOTHER_PASSWORD", "my_secret_key"),
                ("b2b/worker/config.env/SECRET", "super_secret"),
            ],
        ),
    ],
)
def test_get_kv_secrets_v1(vault_api_client, vault_path, expected_results):
    """
    GIVEN a Vault instance and an API client
    WHEN calling get_kv_secrets on a v1 mount
    THEN I get all the expected secrets
    """
    kv_mount = VaultKvMount(name="secret_v1", version="1")

    with my_vcr.use_cassette(
        f"test_hmsl_secret_manager_hashicorp_vault_get_kv_secrets_v1_{vault_path.replace('/', '_')}.yaml",
        ignore_localhost=False,
    ):
        secrets = vault_api_client.get_kv_secrets(kv_mount, vault_path)

    assert secrets == expected_results


@pytest.mark.parametrize(
    "vault_path, expected_results",
    [
        (
            "b2b/web_app/prod/config.env",
            [("b2b/web_app/prod/config.env/PROD_STUFF", "test")],
        ),
        (
            "b2b/worker/config.env",
            [
                ("b2b/worker/config.env/ANOTHER_PASSWORD", "my_secret_key"),
                ("b2b/worker/config.env/SECRET", "super_secret"),
            ],
        ),
    ],
)
def test_get_kv_secrets_v2(vault_api_client, vault_path, expected_results):
    """
    GIVEN a Vault instance and an API client
    WHEN calling get_kv_secrets on a v2 mount
    THEN I get all the expected secrets
    """
    kv_mount = VaultKvMount(name="secret", version="2")

    with my_vcr.use_cassette(
        f"test_hmsl_secret_manager_hashicorp_vault_get_kv_secrets_v2_{vault_path.replace('/', '_')}.yaml",
        ignore_localhost=False,
    ):
        secrets = vault_api_client.get_kv_secrets(kv_mount, vault_path)

    assert secrets == expected_results


@my_vcr.use_cassette(
    "test_hmsl_secret_manager_hashicorp_vault_get_kv_secrets_not_found.yaml",
    ignore_localhost=False,
)
def test_get_kv_secrets_not_found(vault_api_client):
    """
    GIVEN a Vault instance and an API client
    WHEN calling get_kv_secrets with a path that doesn't exist
    THEN VaultNotFoundItemError is raised
    """
    kv_mount = VaultKvMount(name="secret", version="2")

    with pytest.raises(VaultNotFoundItemError):
        vault_api_client.get_kv_secrets(
            kv_mount, "these_are_not_the_secrets_you_are_looking_for.env"
        )


@my_vcr.use_cassette(
    "test_hmsl_secret_manager_hashicorp_vault_get_kv_secrets_forbidden.yaml",
    ignore_localhost=False,
)
def test_get_kv_secrets_forbidden(restricted_vault_api_client):
    """
    GIVEN a Vault instance and an API client
    WHEN calling get_kv_secrets on a path you don't have access to
    THEN VaultForbiddenItemError is raised
    """
    kv_mount = VaultKvMount(name="secret", version="2")

    with pytest.raises(VaultForbiddenItemError):
        restricted_vault_api_client.get_kv_secrets(kv_mount, "b2c/web_app/config.env")


def test_get_secrets_or_empty_success():
    """
    GIVEN a valid mount and path
    WHEN calling _get_secrets_or_empty
    THEN the secrets are returned
    """

    secrets = [("PASSWORD", "my_password"), ("SECRET_KEY", "my_secret_key")]
    with patch(
        "ggshield.verticals.hmsl.secret_manager.hashicorp_vault.api_client."
        "VaultAPIClient.get_kv_secrets",
        return_value=secrets,
    ):
        api_client = VaultAPIClient("vault_url", "vault_token")
        ret = api_client._get_secrets_or_empty(
            VaultKvMount(name="mount_name", version="2"), "my_path"
        )

        assert ret.secrets == secrets
        assert ret.not_fetched_paths == []


def test_get_secrets_or_empty_vault_not_found_item_error():
    """
    GIVEN a valid mount with a path that doesn't exist
    WHEN calling _get_secrets_or_empty
    THEN the inner exception is catched and no secrets are returned
    """

    with patch(
        "ggshield.verticals.hmsl.secret_manager.hashicorp_vault.api_client."
        "VaultAPIClient.get_kv_secrets",
        side_effect=VaultNotFoundItemError,
    ):
        api_client = VaultAPIClient("vault_url", "vault_token")
        ret = api_client._get_secrets_or_empty(
            VaultKvMount(name="mount_name", version="2"), "my_path"
        )

        assert ret.secrets == []
        assert ret.not_fetched_paths == ["my_path"]


def test_get_secrets_or_empty_vault_forbidden_item_error():
    """
    GIVEN a valid mount with a forbidden path
    WHEN calling _get_secrets_or_empty
    THEN the inner exception is catched and no secrets are returned
    """

    with patch(
        "ggshield.verticals.hmsl.secret_manager.hashicorp_vault.api_client."
        "VaultAPIClient.get_kv_secrets",
        side_effect=VaultForbiddenItemError,
    ):
        api_client = VaultAPIClient("vault_url", "vault_token")
        ret = api_client._get_secrets_or_empty(
            VaultKvMount(name="mount_name", version="2"), "my_path"
        )

        assert ret.secrets == []
        assert ret.not_fetched_paths == ["my_path"]


@pytest.mark.parametrize("recursive", [True, False])
def test_get_secrets_from_path_on_directory(recursive):
    """
    GIVEN a valid mount and path of a directory
    WHEN calling _get_secrets_from_path
    THEN the expected results are returned depending on if recursive was set
    """

    directories_res = ["prod", "dev"]
    api_client = VaultAPIClient("vault_url", "vault_token")
    mount = VaultKvMount(name="mount_name", version="2")

    # It's a recursive function, so copy the non-mocked function
    # first for the initial call
    api_client._get_secrets_from_path_not_mocked = api_client._get_secrets_from_path
    with patch(
        "ggshield.verticals.hmsl.secret_manager.hashicorp_vault.api_client."
        "VaultAPIClient._get_secrets_from_path",
    ):
        with patch(
            "ggshield.verticals.hmsl.secret_manager.hashicorp_vault.api_client."
            "VaultAPIClient.list_kv_items",
            return_value=directories_res,
        ):
            ret = api_client._get_secrets_from_path_not_mocked(
                mount, "my_path", recursive
            )

            if not recursive:
                assert ret == VaultSecrets(secrets=[], not_fetched_paths=[])
            else:
                api_client.list_kv_items.assert_called_once_with(mount, "my_path")
                assert api_client._get_secrets_from_path.call_count == 2
                assert api_client._get_secrets_from_path.call_args_list[0][0] == (
                    mount,
                    "my_path/prod",
                    True,
                )
                assert api_client._get_secrets_from_path.call_args_list[1][0] == (
                    VaultKvMount(name="mount_name", version="2"),
                    "my_path/dev",
                    True,
                )


@pytest.mark.parametrize("recursive", [True, False])
def test_get_secrets_from_path_on_file(recursive):
    """
    GIVEN a valid mount and path of a file
    WHEN calling _get_secrets_from_path
    THEN the expected results are returned and recursive parameter has no effect
    """

    secrets_res = VaultSecrets(secrets=[("PASSWORD", "test")], not_fetched_paths=[])
    api_client = VaultAPIClient("vault_url", "vault_token")
    mount = VaultKvMount(name="mount_name", version="2")
    vault_path = "my_path"

    with patch(
        "ggshield.verticals.hmsl.secret_manager.hashicorp_vault.api_client."
        "VaultAPIClient.list_kv_items",
        side_effect=VaultPathIsNotADirectoryError,
    ):
        with patch(
            "ggshield.verticals.hmsl.secret_manager.hashicorp_vault.api_client."
            "VaultAPIClient._get_secrets_or_empty",
            return_value=secrets_res,
        ):
            returned_secrets = api_client._get_secrets_from_path(
                mount, vault_path, recursive
            )

            api_client.list_kv_items.assert_called_once_with(mount, vault_path)
            api_client._get_secrets_or_empty.assert_called_once_with(mount, vault_path)
            assert returned_secrets == secrets_res


@pytest.mark.parametrize("recursive", [True, False])
def test_get_secrets_on_file(recursive):
    """
    GIVEN a valid mount and path of a file
    WHEN calling get_secrets
    THEN we directly return with the secrets of the file and recursive parameter has no effect
    """

    secrets = [
        ("DATABASE_PASSWORD", "my_password"),
        ("ADMIN_PASSWORD", "another_password"),
    ]
    mount = VaultKvMount(name="test", version="2")
    api_client = VaultAPIClient("my_vault_url", "my_vault_token")
    vault_path = "dev/env"

    with patch(
        "ggshield.verticals.hmsl.secret_manager.hashicorp_vault.api_client."
        "VaultAPIClient.list_kv_items",
        side_effect=VaultPathIsNotADirectoryError,
    ):
        with patch(
            "ggshield.verticals.hmsl.secret_manager.hashicorp_vault.api_client."
            "VaultAPIClient._get_secrets_or_empty",
            return_value=VaultSecrets(secrets=secrets, not_fetched_paths=[]),
        ):
            returned_secrets = api_client.get_secrets(mount, vault_path, recursive)
            api_client._get_secrets_or_empty.assert_called_once_with(mount, vault_path)

            assert returned_secrets.secrets == secrets
            assert returned_secrets.not_fetched_paths == []


@pytest.mark.parametrize("recursive", [True, False])
def test_get_secrets_on_directory(recursive):
    """
    GIVEN a valid mount and path of a directory
    WHEN calling get_secrets
    THEN _get_secrets_from_path is called with the correct parameters
    """

    directories = ["dev", "prod", "sandbox"]
    mount = VaultKvMount(name="test", version="2")
    api_client = VaultAPIClient("my_vault_url", "my_vault_token")
    vault_path = ""

    with patch(
        "ggshield.verticals.hmsl.secret_manager.hashicorp_vault.api_client."
        "VaultAPIClient.list_kv_items",
        return_value=directories,
    ):
        with patch(
            "ggshield.verticals.hmsl.secret_manager.hashicorp_vault.api_client."
            "VaultAPIClient._get_secrets_from_path",
            side_effect=lambda _, called_path, __: VaultSecrets(
                secrets=[
                    (
                        f"PASSWORD_{called_path}",
                        "my_password",
                    )
                ],
                not_fetched_paths=[],
            ),
        ):
            returned_secrets = api_client.get_secrets(mount, vault_path, recursive)

            api_client.list_kv_items.assert_called_once_with(mount, vault_path)
            assert api_client._get_secrets_from_path.call_count == 3

            assert api_client._get_secrets_from_path.call_args_list[0][0] == (
                mount,
                "dev",
                recursive,
            )
            assert api_client._get_secrets_from_path.call_args_list[1][0] == (
                mount,
                "prod",
                recursive,
            )
            assert api_client._get_secrets_from_path.call_args_list[2][0] == (
                mount,
                "sandbox",
                recursive,
            )
            assert returned_secrets.secrets == [
                ("PASSWORD_dev", "my_password"),
                ("PASSWORD_prod", "my_password"),
                ("PASSWORD_sandbox", "my_password"),
            ]



import shutil
from io import StringIO
from typing import Dict, List, Optional, Tuple

from pygitguardian.client import VERSIONS

from ggshield.core.constants import IncidentStatus
from ggshield.core.lines import Line, get_offset, get_padding
from ggshield.core.text_utils import (
    STYLE,
    clip_long_line,
    format_text,
    pluralize,
    translate_validity,
)

from ..extended_match import ExtendedMatch
from ..secret_scan_collection import (
    IgnoreKind,
    Result,
    Secret,
    SecretScanCollection,
    group_secrets_by_ignore_sha,
)
from .secret_output_handler import SecretOutputHandler


# MAX_SECRET_SIZE controls the max length of |-----| under a secret
# avoids occupying a lot of space in a CI terminal.
MAX_SECRET_SIZE = 80
# The number of lines to display before and after a secret in the patch
NB_CONTEXT_LINES = 3


def secret_file_info(
    filename: str,
    incident_count: int,
    incident_status: IncidentStatus = IncidentStatus.DETECTED,
    hidden_secrets_count: int = 0,
) -> str:
    """Return the formatted file info (number of secrets, filename, optional number of hidden secrets)."""
    result = "\n{} {}: {} {} {}\n".format(
        format_text(">", STYLE["detector_line_start"]),
        format_text(filename, STYLE["filename"]),
        incident_count,
        pluralize("secret", incident_count, "secrets"),
        incident_status.value,
    )
    if hidden_secrets_count > 0:
        result += "{} {}  {} {} ignored - use the `--all-secrets` option to display {}\n".format(
            format_text(">", STYLE["detector_line_start"]),
            " " * len(filename),
            hidden_secrets_count,
            pluralize("secret", hidden_secrets_count, "secrets"),
            pluralize("it", hidden_secrets_count, "them"),
        )
    return result


class SecretTextOutputHandler(SecretOutputHandler):
    def _process_scan_impl(self, scan: SecretScanCollection) -> str:
        """Output Secret Scan Collection in text format"""
        processed_scan_results = self.process_scan_results(scan)

        scan_buf = StringIO()
        if self.verbose:
            scan_buf.write(secrets_engine_version())
        scan_buf.write(processed_scan_results)
        if not processed_scan_results:

            scan_buf.write(
                no_new_leak_message()
                if self.ignore_known_secrets
                else no_leak_message()
            )

        known_secrets_count = sum(
            result.ignored_secrets_count_by_kind.get(IgnoreKind.KNOWN_SECRET, 0)
            for result in scan.get_all_results()
        )
        if self.ignore_known_secrets and known_secrets_count > 0:
            scan_buf.write(
                f"\nWarning: {known_secrets_count} {pluralize('secret', known_secrets_count)} ignored "
                f"because {pluralize('it is', known_secrets_count, 'they are')} already known by your "
                f"GitGuardian dashboard and you used the `--ignore-known-secrets` option.\n"
            )
            # TODO: display that using --all-secrets will display those
        return scan_buf.getvalue()

    def process_scan_results(self, scan: SecretScanCollection) -> str:
        """Iterate through the scans and sub-scan results to prepare the display."""
        results_buf = StringIO()
        if scan.results:
            current_result_buf = StringIO()
            for result in scan.results.results:
                current_result_buf.write(self.process_result(result))
            current_result_string = current_result_buf.getvalue()

            # We want to show header when at least one result is not empty
            if scan.optional_header and current_result_string:
                results_buf.write(scan.optional_header)

            results_buf.write(current_result_string)

        if scan.scans:
            for sub_scan in scan.scans:
                inner_scan_str = self.process_scan_results(sub_scan)
                results_buf.write(inner_scan_str)

        return results_buf.getvalue()

    def process_result(self, result: Result) -> str:
        """
        Build readable message on the found incidents.

        :param result: The result from scanning API
        :param show_secrets: Option to show secrets value
        :param show_only_known_secrets: If True, display only known secrets, and only new secrets otherwise
        :return: The formatted message to display
        """
        result_buf = StringIO()

        sha_dict = group_secrets_by_ignore_sha(result.secrets)

        if not self.show_secrets:
            result.censor()

        number_of_displayed_secrets = 0
        number_of_hidden_secrets = sum(result.ignored_secrets_count_by_kind.values())
        for ignore_sha, secrets in sha_dict.items():
            number_of_displayed_secrets += 1

            result_buf.write(
                secret_header(secrets, ignore_sha, secrets[0].known_secret)
            )

            result_buf.write(
                leak_message_located(
                    flatten_secrets_by_line(secrets),
                    result.is_on_patch,
                    clip_long_lines=not self.verbose,
                )
            )

        file_info_line = ""
        if number_of_displayed_secrets > 0 or number_of_hidden_secrets > 0:
            file_info_line = secret_file_info(
                result.filename,
                incident_count=number_of_displayed_secrets,
                hidden_secrets_count=number_of_hidden_secrets,
            )

        return file_info_line + result_buf.getvalue()


def leak_message_located(
    flat_matches: List[Tuple[Line, List[ExtendedMatch]]],
    is_patch: bool = False,
    clip_long_lines: bool = False,
) -> str:
    """
    Display leak message of an incident with location in content.

    :param flat_matches:  a list of tuples mapping a line to a list of matches starting at that line.
    :param padding: The line padding
    :param offset: The offset due to the line display
    :param is_patch: Whether the content is a patch
    :param clip_long_lines: Whether to clip long lines
    """
    leak_msg = StringIO()
    all_lines = []
    for line, list_matches in flat_matches:
        if not list_matches:
            all_lines.append(line)
        for match in list_matches:
            all_lines.extend(
                match.lines_with_secret,
            )
    padding = get_padding(all_lines)
    offset = get_offset(padding, is_patch)
    max_width = shutil.get_terminal_size()[0] - offset if clip_long_lines else 0

    old_line_number: Optional[int] = None
    for line, matches in flat_matches:
        line_number = line.pre_index or line.post_index
        line_content = line.content

        if (
            line_number is not None
            and old_line_number is not None
            and line_number - old_line_number != 1
        ):
            leak_msg.write(format_line_count_break(padding))

        if not matches:
            # The current line is just here for context
            leak_msg.write(line.build_line_count(padding, is_secret=False))
            if clip_long_lines:
                line_content = clip_long_line(
                    line_content, max_width - int(is_patch), after=True
                )
            leak_msg.write(f"{display_patch(line_content)}\n")
        else:
            # The current line number matches a found secret
            assert line_number is not None  # we cannot have secret in patch hunks
            for match in matches:
                span = match.span
                if len(match.lines_with_secret) == 1:
                    # The secret is on just one line
                    leak_msg.write(line.build_line_count(padding, is_secret=True))
                    formatted_line, detector_position = format_line_with_secret(
                        line_content,
                        span.column_index_start,
                        span.column_index_end,
                        max_width,
                        is_patch,
                    )
                    leak_msg.write(formatted_line)
                else:
                    # The secret is on multiple lines
                    detector_position = float("inf"), float("-inf")
                    for index, line_of_secret in enumerate(match.lines_with_secret):
                        leak_msg.write(
                            line_of_secret.build_line_count(padding, is_secret=True)
                        )
                        secret_start = span.column_index_start if index == 0 else 0
                        secret_end = (
                            span.column_index_end
                            if index == len(match.lines_with_secret) - 1
                            else len(line_of_secret.content)
                        )
                        formatted_line, secret_position = format_line_with_secret(
                            line_of_secret.content,
                            secret_start,
                            secret_end,
                            max_width,
                            is_patch,
                        )
                        leak_msg.write(formatted_line)

                        detector_position = (
                            min(detector_position[0], secret_position[0]),
                            max(detector_position[1], secret_position[1]),
                        )
                    line_number += len(match.lines_with_secret) - 1
                detector_position = int(detector_position[0]), int(detector_position[1])
                detector = format_detector(match.match_type, *detector_position)
                leak_msg.write(display_detector(detector, offset))

        old_line_number = line_number

    return leak_msg.getvalue()


def flatten_secrets_by_line(
    secrets: List[Secret],
) -> List[Tuple[Line, List[ExtendedMatch]]]:
    """
    flatten_secrets_by_line turns a list of policy breaks into a list of
    tuples of (line, list of matches) mapping a line to a list of matches starting
    at that line. The list is sorted by line number.
    """
    flat_match_dict: Dict[Line, List[ExtendedMatch]] = dict()
    for secret in secrets:
        for match in secret.matches:
            assert isinstance(match, ExtendedMatch)
            for line in match.lines_before_secret + match.lines_after_secret:
                if line not in flat_match_dict:
                    flat_match_dict[line] = []
            # Only add match to the first line, we will handle multiline at formating
            line = match.lines_with_secret[0]
            if line not in flat_match_dict:
                flat_match_dict[line] = []
            flat_match_dict[line].append(match)
    # Sort the matches per line number
    ordered_flat_match: List[Tuple[Line, List[ExtendedMatch]]] = sorted(
        flat_match_dict.items(),
        key=lambda item: item[0].pre_index or item[0].post_index or -1,
    )
    return ordered_flat_match


def secret_header(
    secrets: List[Secret],
    ignore_sha: str,
    known_secret: bool = False,
) -> str:
    """
    Build a header for the policy break.
    """
    secret = secrets[0]
    indent = "   "
    validity_msg = (
        f"\n{indent}Validity: {format_text(translate_validity(secret.validity), STYLE['incident_validity'])}"
        if secret.validity
        else ""
    )

    start_line = format_text(">>", STYLE["detector_line_start"])
    secret_type = format_text(secret.detector, STYLE["secret_type"])
    number_occurrences = format_text(str(len(secrets)), STYLE["occurrence_count"])
    ignore_sha = format_text(ignore_sha, STYLE["ignore_sha"])

    message = f"""
{start_line} Secret detected: {secret_type}{validity_msg}
{indent}Occurrences: {number_occurrences}
{indent}Known by GitGuardian dashboard: {"YES" if known_secret else "NO"}
{indent}Incident URL: {secrets[0].incident_url if known_secret and secret.incident_url else "N/A"}
{indent}Secret SHA: {ignore_sha}
"""
    if secret.ignore_reason is not None:
        message += f"{indent}Ignored: {secret.ignore_reason.to_human_readable()}\n"

    return message + "\n"


def no_leak_message() -> str:
    """
    Build a message if no secret is found.
    """
    return format_text("\nNo secrets have been found\n", STYLE["no_secret"])


def no_new_leak_message() -> str:
    """
    Build a message if no new secret is found.
    """
    return format_text("\nNo new secrets have been found\n", STYLE["no_secret"])


def format_line_with_secret(
    line_content: str,
    secret_index_start: int,
    secret_index_end: int,
    max_width: Optional[int] = None,
    is_patch: bool = False,
) -> Tuple[str, Tuple[int, int]]:
    """
    Format a line containing a secret.
    :param line_content: the whole line, as a string
    :param secret_index_start: the index in the line, as an integer, where the secret
    starts
    :param secret_index_end: the index in the line, as an integer, where the secret
    ends
    :param max_width: if set, context will be clipped if needed
    :return The formatted detector as a string, and the position (start and end index)
    of the secret in it.
    """
    context_before = line_content[:secret_index_start]
    secret = line_content[secret_index_start:secret_index_end]
    context_after = line_content[secret_index_end:]
    secret_length = len(secret)

    # Clip the context if it is too long
    if max_width:
        context_max_length = max_width - secret_length
        if len(context_before) + len(context_after) > context_max_length:
            # Both before and after context are too long, cut them to the same size
            if (
                len(context_before) > context_max_length // 2
                and len(context_after) > context_max_length // 2
            ):
                context_before = clip_long_line(
                    context_before,
                    context_max_length // 2,
                    before=True,
                    is_patch=is_patch,
                )
                context_after = clip_long_line(
                    context_after, context_max_length // 2, after=True
                )
            # Only the before context is too long, clip it but use the maximum space
            # available
            elif len(context_before) > context_max_length // 2:
                context_before = clip_long_line(
                    context_before,
                    context_max_length - len(context_after),
                    before=True,
                    is_patch=is_patch,
                )
            # Only the after context is too long, same idea
            elif len(context_after) > context_max_length // 2:
                context_after = clip_long_line(
                    context_after, context_max_length - len(context_before), after=True
                )

    formatted_line = (
        (display_patch(context_before) if context_before else "")
        + display_match_value(secret)
        + (display_patch(context_after) if context_after else "")
        + "\n"
    )

    secret_display_index_start = len(context_before)
    secret_display_index_end = secret_display_index_start + secret_length
    return formatted_line, (secret_display_index_start, secret_display_index_end)


def display_patch(patch: str) -> str:
    """Return the formatted patch."""
    return format_text(patch, STYLE["patch"])


def display_match_value(match_value: str) -> str:
    """Return the formatted match value."""
    return format_text(match_value, STYLE["secret"])


def display_detector(detector: str, offset: int) -> str:
    """Return the formatted detector line."""
    return " " * offset + format_text(detector, STYLE["detector"])


def format_detector(match_type: str, index_start: int, index_end: int) -> str:
    """Return detector object to add in detector_line."""

    detector_size = len(match_type)
    secret_size = index_end - index_start

    display = ""
    if secret_size < MAX_SECRET_SIZE:
        before = "_" * max(1, int(((secret_size - detector_size) - 1) / 2))
        after = "_" * max(1, (secret_size - len(before) - detector_size) - 2)
        display = f"|{before}{match_type}{after}|"

    return " " * index_start + format_text(display, STYLE["detector"]) + "\n"


def secrets_engine_version() -> str:
    return f"\nsecrets-engine-version: {VERSIONS.secrets_engine_version}\n"


def format_line_count_break(padding: int) -> str:
    """Return the line count break."""
    return format_text(
        " " * max(0, padding - len("...")) + "...\n", STYLE["detector_line_start"]
    )



import re
from dataclasses import dataclass
from typing import Iterable, List, Optional

from ggshield.core.errors import ParseError
from ggshield.core.text_utils import STYLE, format_line_count, format_text
from ggshield.utils.git_shell import Filemode


REGEX_HUNK_HEADER = re.compile(
    r"^@@ -(?P<pre_index>\d+),?\d* \+(?P<post_index>\d+),?\d* @@(?P<trailing_content> .+)?"  # noqa
)

# Prefix used before each line for a complete file
FILE_LINE_PREFIX = "{} | "

# Prefix used before each line for a patch
PATCH_LINE_PREFIX = "{} {} | "


@dataclass
class Line:
    """
    Represent a line in a document.

    - content: Content of the line
    - pre_index: Line index (deletion for patches, line index for files)
    - post_index: Line index (addition for patches)

    pre_index and post_index are 1-based.

    For files: pre_index is the line number, post_index is None.

    For patches: pre_index and post_index are line numbers relative to the *document
    modified by the patch*, not to the patch itself. This means that for a patch like
    this:

    ```
    diff --git a/foo.py b/foo.py
    index c53e550a..7ed9d026 100644
    --- a/foo.py
    +++ b/foo.py
    @@ -56,8 +56,8 @@
     Hello
    -world
    +beautiful
    +WORLD
    ```

    The values of content, pre_index, and post_index are:

    content                 pre_index  post_index
    "@@ -56,8 +56,8 @@"     None        None
    " Hello"                56          56
    "-world"                57          None
    "+beautiful"            None        57
    "+WORLD"                None        58
    """

    content: str
    is_patch: bool = False
    pre_index: Optional[int] = None
    post_index: Optional[int] = None

    def __hash__(self) -> int:
        return hash(
            (
                self.content,
                self.is_patch,
                self.pre_index,
                self.post_index,
            )
        )

    def build_line_count(self, padding: int, is_secret: bool = False) -> str:
        """Return the formatted line count."""
        line_count_style = (
            STYLE["line_count_secret"] if is_secret else STYLE["line_count"]
        )

        # File
        if not self.is_patch:
            return FILE_LINE_PREFIX.format(
                format_text(
                    format_line_count(self.pre_index, padding), line_count_style
                )
            )

        # Patch
        pre_index = format_line_count(self.pre_index, padding)
        post_index = format_line_count(self.post_index, padding)

        return PATCH_LINE_PREFIX.format(
            format_text(pre_index, line_count_style),
            format_text(post_index, line_count_style),
        )


def get_lines_from_content(content: str, filemode: Filemode) -> List[Line]:
    """
    Return the secrets and the lines with line number.

    :param content: Content to scan
    :param filemode: Filemode of the content
    """

    # Patch
    if filemode != Filemode.FILE:
        return list(get_lines_from_patch(content, filemode))

    # File
    return list(get_lines_from_file(content))


def get_lines_from_file(content: str) -> Iterable[Line]:
    """Return the lines with line number from a file."""
    for line_count, line_content in enumerate(content.split("\n")):
        yield Line(content=line_content, is_patch=False, pre_index=line_count + 1)


def get_lines_from_patch(content: str, filemode: Filemode) -> Iterable[Line]:
    """Return the lines with line number from a git patch."""
    content += "\n"
    pre_index = 0
    post_index = 0

    for line in content.split("\n"):
        if not line:
            continue
        line_type = line[0]
        line_content = line
        line_pre_index = None
        line_post_index = None

        if line_type == " ":
            pre_index += 1
            post_index += 1
            line_pre_index = pre_index
            line_post_index = post_index
        elif line_type == "@":
            m = REGEX_HUNK_HEADER.search(line)
            if m is None:
                continue
            pre_index = int(m.groupdict()["pre_index"])
            post_index = int(m.groupdict()["post_index"])
            trailing_content = m.groupdict()["trailing_content"]

            if filemode == Filemode.NEW or filemode == Filemode.DELETE:
                pre_index = 1
                post_index = 1

            pre_index -= 1
            post_index -= 1
            if trailing_content:
                # If the hunk header has trailing content, it might contain a secret
                # match, so give it a line number.
                #
                # That trailing content is generated by Git (see [1]), so the number
                # might not be correct, but the rest of ggshield requires it.
                #
                # [1]: https://git-scm.com/docs/gitattributes#_defining_a_custom_hunk_header
                line_pre_index = pre_index
                line_post_index = pre_index
        elif line_type == "+":
            post_index += 1
            line_post_index = post_index
        elif line_type == "-":
            pre_index += 1
            line_pre_index = pre_index
        elif line_type == "\\":
            # This type of line shouldn't contain any secret; no need to set indices
            pass
        else:
            raise ParseError(
                f'Failed to parse patch line. Unexpected first character in "{line}"'
            )

        yield Line(
            content=line_content,
            is_patch=True,
            pre_index=line_pre_index,
            post_index=line_post_index,
        )


def get_padding(lines: List[Line]) -> int:
    """Return the number of digit of the maximum line number."""
    return max(len(str(lines[-1].pre_index or 0)), len(str(lines[-1].post_index or 0)))


def get_offset(padding: int, is_patch: bool = False) -> int:
    """Return the offset due to the line prefix."""
    if is_patch:
        return len(PATCH_LINE_PREFIX.format("0" * padding, "0" * padding))

    return len(FILE_LINE_PREFIX.format("0" * padding))



from typing import List, Optional, Union
from unittest import mock

import pytest

from ggshield import __version__
from ggshield.verticals.hmsl import HMSLClient
from ggshield.verticals.hmsl.client import PREFIX_LENGTH
from ggshield.verticals.hmsl.crypto import hash_string


SAMPLE_QUERY = ["408a5b05c35bb4d230e31da1f9afa0e8881050cb72775e925d6bc7cb945b4f39"]
SAMPLE_RESPONSE = {
    "matches": [
        {
            "hint": "5de9f935ee515de855551e1786d71f1d7f4c3805083f57dc49863ad624f5ba42",
            "payload": (
                "gubW5fk1nZF8I4yGVMdcpeZv3CDg5lKE4LTeWDmjD3YkuaMx9c7/HuzIv8T3"
                "X2KXAcEWME+wQBiUP0WJQY6dEUNbcvyBpuxl+KGLj7KC3Dxlt8U+Frv4vDnFwWouN3c="
            ),
        },
        {
            "hint": "df096cd78ff6d840a21d95db852d69c3bb7710945123841fe0cfae88561a32be",
            "payload": (
                "sEdfOFMcwEmt2ZNwRL4nuoiTsnHbycdC/OOUPOM0ktZQhQQmG60+Js57T003"
                "SP6brRO4EZILO9cWJGAaCWiv21rAgwCt2r6IZn2KodRq0S4RWHajbphRgOvw3j6oPQ=="
            ),
        },
    ]
}


@pytest.fixture
def hashes():
    """Return example hashes to serve as payload."""
    return [hash_string(letter) for letter in "abcdefghijklmnopqrstuvwxyz"]


def get_hmsl_client(
    hmsl_command_path: Optional[str] = None,
    prefix_length: int = PREFIX_LENGTH,
    jwt: Union[str, None] = None,
):
    """Return a HMSL client."""
    return HMSLClient(
        "foo",
        hmsl_command_path or "ggshield hmsl check",
        jwt=jwt,
        prefix_length=prefix_length,
    )


def test_hmsl_client_check_prefixes(hashes: List[str]):
    """
    GIVEN a HMSL client
    WHEN a batch of hashes is audited
    THEN the query contains prefixes
    """
    hmsl_client = get_hmsl_client(prefix_length=6)
    expected_prefixes = {hash[:6] for hash in hashes[:5]}

    with mock.patch("requests.Session.post") as post:
        list(hmsl_client.check(hashes[:5]))
        url, kwargs = post.call_args
        assert url == ("foo/v1/prefixes",)
        assert (
            "prefixes" in kwargs["json"]
            and set(kwargs["json"]["prefixes"]) == expected_prefixes
        )


def test_hmsl_client_check_hashes(hashes):
    """
    GIVEN a HMSL client
    WHEN a batch of hashes is audited as full hashes
    THEN the query contains full hashes
    """
    hmsl_client = get_hmsl_client()

    with mock.patch("requests.Session.post") as post:
        list(hmsl_client.check(hashes + hashes, full_hashes=True))
        url, kwargs = post.call_args
        assert url == ("foo/v1/hashes",)
        assert "hashes" in kwargs["json"] and set(kwargs["json"]["hashes"]) == set(
            hashes
        )


def test_hmsl_client_jwt(hashes):
    """
    GIVEN a HMSL client
    WHEN a JWT is provided
    THEN the query contains it as an Bearer token
    """
    hmsl_client = get_hmsl_client(jwt="bar")

    with mock.patch("requests.Session.post") as post:
        list(hmsl_client.check([]))
        assert not post.called
        list(hmsl_client.check(hashes[:1]))
        _, kwargs = post.call_args
        assert ("Authorization", "Bearer bar") in kwargs["headers"].items()


def test_hmsl_client_decode_response():
    """
    GIVEN a HMSL client
    WHEN a response is received
    THEN the response is properly decrypted
    """
    hmsl_client = get_hmsl_client()

    with mock.patch(
        "ggshield.verticals.hmsl.HMSLClient._query", return_value=SAMPLE_RESPONSE
    ):
        response = list(hmsl_client.check(SAMPLE_QUERY, full_hashes=False))
    assert len(response) == 1
    secret = response[0]
    assert secret.count == 42
    assert "github.com" in secret.url


@pytest.mark.parametrize(
    "command, expected_header_value",
    [
        (
            "ggshield hmsl check-secret-manager hashicorp-vault",
            "hmsl_check-secret-manager_hashicorp-vault",
        ),
        ("ggshield hmsl check", "hmsl_check"),
    ],
)
def test_hmsl_common_headers(command, expected_header_value):
    """
    GIVEN the HMSLClient class
    WHEN creating a new client instance
    THEN the underlying requests session has the GGShield-HMSL-Command-Name header set correctly
    """

    hmsl_client = get_hmsl_client(command)

    assert (
        hmsl_client.session.headers["GGShield-HMSL-Command-Name"]
        == expected_header_value
    )
    assert hmsl_client.session.headers["User-Agent"] == f"GGShield {__version__}"



from pathlib import Path
from typing import Callable, Iterable, Optional, Pattern, Sequence, Set

from ggshield.utils.git_shell import git
from ggshield.utils.itertools import batched
from ggshield.utils.os import getenv_int

from .commit_information import CommitInformation
from .commit_utils import (
    DIFF_EMPTY_COMMIT_INFO_BLOCK,
    MAX_FILES_PER_GIT_COMMAND,
    PATCH_COMMON_ARGS,
    PATCH_PREFIX,
    STAGED_PREFIX,
    CommitScannable,
    get_diff_files,
    get_file_sha_in_ref,
    get_file_sha_stage,
    parse_patch,
)
from .scannable import Scannable


# get files in commit by batch of _MAX_DOCS_PER_COMMIT
_MAX_DOCS_PER_COMMIT = getenv_int("GG_MAX_DOCS_PER_COMMIT", 20)

# Internal: type for a function to produce scannables from a commit
PatchParserFunction = Callable[["Commit"], Iterable[Scannable]]


class Commit:
    """
    Commit represents a commit which is a list of commit files.
    """

    def __init__(
        self,
        sha: Optional[str],
        patch_parser: PatchParserFunction,
        info: CommitInformation,
    ):
        """
        Internal constructor. Used by the `from_*` static methods and by some tests.
        Real code should use the `from_*` methods.
        """
        self.sha = sha
        self._patch_parser = patch_parser
        self.info = info

    @property
    def urls(self) -> Sequence[str]:
        return [CommitScannable.create_url(self.sha, x) for x in self.info.paths]

    @staticmethod
    def from_sha(
        sha: str,
        exclusion_regexes: Optional[Set[Pattern[str]]] = None,
        cwd: Optional[Path] = None,
    ) -> "Commit":
        info = CommitInformation.from_sha(sha, cwd=cwd)

        def parser(commit: "Commit") -> Iterable[Scannable]:
            for paths in batched(commit.info.paths, _MAX_DOCS_PER_COMMIT):
                cmd = ["show", sha, *PATCH_COMMON_ARGS, "--"]

                # Append paths to the command-line. If the file has been renamed, append
                # both old and new paths. If we only append the new path then `git
                # show` returns the new path as an added file.
                for path in paths:
                    if old_path := commit.info.renames.get(path):
                        cmd.append(str(old_path))
                    cmd.append(str(path))

                patch = git(cmd, cwd=cwd)
                yield from parse_patch(sha, patch, exclusion_regexes)

        return Commit(sha, parser, info)

    @staticmethod
    def from_merge(
        exclusion_regexes: Optional[Set[Pattern[str]]] = None,
        merge_branch: str = "MERGE_HEAD",
        cwd: Optional[Path] = None,
    ) -> "Commit":

        diff_files = get_diff_files(cwd=cwd)

        shas_in_merge_branch = dict(
            get_file_sha_in_ref(merge_branch, diff_files, cwd=cwd)
        )
        shas_in_head = dict(get_file_sha_in_ref("HEAD", diff_files, cwd=cwd))

        files_to_scan = set()
        for path, sha in get_file_sha_stage(diff_files, cwd=cwd):
            # The file is either new or different from both HEAD and MERGE_HEAD
            if sha not in {shas_in_head.get(path), shas_in_merge_branch.get(path)}:
                files_to_scan.add(path)

        def parser_merge(commit: "Commit") -> Iterable[Scannable]:
            for files in batched(files_to_scan, MAX_FILES_PER_GIT_COMMAND):
                patch = git(["diff", "--staged", *PATCH_COMMON_ARGS, *files], cwd=cwd)
                yield from parse_patch(
                    STAGED_PREFIX,
                    DIFF_EMPTY_COMMIT_INFO_BLOCK + patch,
                    exclusion_regexes,
                )

        info = CommitInformation.from_staged(cwd)
        return Commit(sha=None, patch_parser=parser_merge, info=info)

    @staticmethod
    def from_staged(
        exclusion_regexes: Optional[Set[Pattern[str]]] = None,
        cwd: Optional[Path] = None,
    ) -> "Commit":
        def parser(commit: "Commit") -> Iterable[Scannable]:
            patch = git(["diff", "--staged"] + PATCH_COMMON_ARGS, cwd=cwd)
            yield from parse_patch(
                STAGED_PREFIX,
                DIFF_EMPTY_COMMIT_INFO_BLOCK + patch,
                exclusion_regexes,
            )

        info = CommitInformation.from_staged(cwd)

        return Commit(sha=None, patch_parser=parser, info=info)

    @staticmethod
    def from_patch(
        patch: str,
        exclusion_regexes: Optional[Set[Pattern[str]]] = None,
        sha: str = PATCH_PREFIX,
    ) -> "Commit":
        """This one is for tests"""
        info = CommitInformation.from_patch_header(patch)

        def parser(commit: "Commit") -> Iterable[Scannable]:
            yield from parse_patch(sha, patch, exclusion_regexes)

        return Commit(sha=None, patch_parser=parser, info=info)

    def get_files(self) -> Iterable[Scannable]:
        """
        Parse the patch into files and extract the changes for each one of them.
        """
        yield from self._patch_parser(self)

    def __repr__(self) -> str:
        return f"<Commit sha={self.sha}>"



from typing import List

import pytest
from pygitguardian.models import ScanResult

from ggshield.core.lines import get_lines_from_content
from ggshield.core.match_span import MatchSpan
from ggshield.utils.git_shell import Filemode
from tests.unit.conftest import (
    _MULTI_SECRET_ONE_LINE_PATCH,
    _MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT,
    _MULTI_SECRET_TWO_LINES_PATCH,
    _MULTI_SECRET_TWO_LINES_PATCH_SCAN_RESULT,
    _ONE_LINE_AND_MULTILINE_PATCH_CONTENT,
    _ONE_LINE_AND_MULTILINE_PATCH_SCAN_RESULT,
    _SIMPLE_SECRET_MULTILINE_PATCH,
    _SIMPLE_SECRET_MULTILINE_PATCH_SCAN_RESULT,
    _SIMPLE_SECRET_PATCH,
    _SIMPLE_SECRET_PATCH_SCAN_RESULT,
)


@pytest.mark.parametrize(
    ["content", "scan_result", "expected_spans"],
    [
        pytest.param(
            _SIMPLE_SECRET_PATCH,
            _SIMPLE_SECRET_PATCH_SCAN_RESULT,
            [MatchSpan(1, 1, 15, 56)],
            id="1match",
        ),
        pytest.param(
            _MULTI_SECRET_ONE_LINE_PATCH,
            _MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT,
            [
                MatchSpan(1, 1, 17, 32),
                MatchSpan(1, 1, 54, 86),
            ],
            id="multimatch-1line",
        ),
        pytest.param(
            _MULTI_SECRET_TWO_LINES_PATCH,
            _MULTI_SECRET_TWO_LINES_PATCH_SCAN_RESULT,
            [
                MatchSpan(1, 1, 17, 32),
                MatchSpan(2, 2, 21, 53),
            ],
            id="multimatch-2lines",
        ),
        pytest.param(
            _SIMPLE_SECRET_MULTILINE_PATCH,
            _SIMPLE_SECRET_MULTILINE_PATCH_SCAN_RESULT,
            [
                MatchSpan(2, 10, 9, 32),
            ],
            id="1match-multiline",
        ),
        pytest.param(
            _ONE_LINE_AND_MULTILINE_PATCH_CONTENT,
            _ONE_LINE_AND_MULTILINE_PATCH_SCAN_RESULT,
            [
                MatchSpan(1, 1, 18, 33),
                MatchSpan(1, 1, 36, 68),
                MatchSpan(1, 9, 69, 30),
                MatchSpan(9, 9, 38, 107),
            ],
            id="1line-1multiline",
        ),
    ],
)
def test_from_match(
    content: str,
    scan_result: ScanResult,
    expected_spans: List[MatchSpan],
):
    """
    GIVEN a patch content
    AND its scan result
    WHEN MatchSpan.from_match() is called on all the result matches
    THEN the created MatchSpans are equal to the expected spans
    """
    lines = get_lines_from_content(content, filemode=Filemode.NEW)

    matches = [
        match
        for policy_break in scan_result.policy_breaks
        for match in policy_break.matches
    ]
    for idx, (expected_span, match) in enumerate(zip(expected_spans, matches)):
        span = MatchSpan.from_match(match, lines)
        assert span == expected_span, f"Error on match[{idx}]"

    assert len(expected_spans) == len(matches)



from dataclasses import dataclass
from enum import Enum, auto
from typing import (
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    TextIO,
    Tuple,
)

from dotenv import dotenv_values

from ggshield.core.filter import censor_string
from ggshield.verticals.hmsl import PREFIX_LENGTH
from ggshield.verticals.hmsl.crypto import hash_string
from ggshield.verticals.hmsl.utils import EXCLUDED_KEYS, EXCLUDED_VALUES


class InputType(Enum):
    FILE = auto()
    ENV = auto()


@dataclass
class PreparedSecrets:
    payload: Set[str]
    mapping: Dict[str, str]


@dataclass
class SecretWithKey:
    key: Optional[str]
    value: str


# Methods to compute names for secrets
# They are useful to help the user identify the secret that may leaked
# with a more "human-readable" string than a hash.
# Takes the secret and optional key as input and returns a string.

NamingStrategy = Callable[[SecretWithKey], str]
NAMING_STRATEGIES: Dict[str, NamingStrategy] = {
    "censored": lambda secret: censor_string(secret.value),
    "cleartext": lambda secret: secret.value,
    "none": lambda _: "",
    "key": lambda secret: secret.key or censor_string(secret.value),
}


def collect_list(input: List[Tuple[str, str]]) -> Iterator[SecretWithKey]:
    """
    Collect the secrets to pass to prepare.

    Input should be a list of tuple with the first item of the tuple being the secret
    key and the second the secret value.
    """
    for key, value in input:
        # filter our excluded keys and values
        if not key or not value:
            continue
        if (
            key.split("/")[-1].upper() in EXCLUDED_KEYS  # Only use the variable name
            or value.lower() in EXCLUDED_VALUES
        ):
            continue
        yield SecretWithKey(key=key, value=value)


def collect(
    input: TextIO, input_type: InputType = InputType.FILE
) -> Iterator[SecretWithKey]:
    """
    Collect the secrets to pass to prepare.
    """
    if input_type == InputType.ENV:
        config = dotenv_values(stream=input)
        for key, value in config.items():
            # filter our excluded keys and values
            if not key or not value:
                continue
            if key.upper() in EXCLUDED_KEYS or value.lower() in EXCLUDED_VALUES:
                continue
            yield SecretWithKey(value=value, key=key)
    else:
        for line in input:
            secret = line.strip()
            if secret == "":
                # Skip empty lines
                continue
            yield SecretWithKey(value=secret, key=None)


def prepare(
    secrets: Iterable[SecretWithKey],
    naming_strategy: NamingStrategy,
    *,
    full_hashes: bool = False,
) -> PreparedSecrets:
    """
    Prepare the secrets so they can later be checked.
    """
    hashes: Set[str] = set()
    mapping: Dict[str, str] = {}
    for secret in secrets:
        name = naming_strategy(secret)
        hash = hash_string(secret.value)
        mapping[hash] = name
        if full_hashes:
            hashes.add(hash)
        else:
            hashes.add(hash[:PREFIX_LENGTH])
    return PreparedSecrets(
        payload=hashes,
        mapping=mapping,
    )



from pathlib import Path
from subprocess import CalledProcessError

import pytest

from tests.functional.utils_create_merge_repo import (
    SecretLocation,
    generate_repo_with_merge_commit,
)


@pytest.mark.parametrize(
    "with_conflict",
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    "secret_location",
    [
        SecretLocation.MASTER_BRANCH,
        SecretLocation.FEATURE_BRANCH,
        SecretLocation.NO_SECRET,
    ],
)
@pytest.mark.parametrize(
    "scan_all_merge_files",
    [
        True,
        False,
    ],
)
def test_merge_commit_no_conflict(
    capsys,
    tmp_path: Path,
    with_conflict: bool,
    secret_location: SecretLocation,
    scan_all_merge_files: bool,
) -> None:

    if (
        secret_location == SecretLocation.MASTER_BRANCH
        and with_conflict
        and scan_all_merge_files
    ):
        with pytest.raises(CalledProcessError):
            generate_repo_with_merge_commit(
                tmp_path,
                with_conflict=with_conflict,
                secret_location=secret_location,
                scan_all_merge_files=scan_all_merge_files,
            )

        # AND the error message contains the Gitlab Token
        captured = capsys.readouterr()
        assert "GitLab Token" in captured.err
    else:
        generate_repo_with_merge_commit(
            tmp_path,
            with_conflict=with_conflict,
            secret_location=secret_location,
            scan_all_merge_files=scan_all_merge_files,
        )


def test_merge_commit_with_conflict_and_secret_in_conflict(
    tmp_path: Path,
) -> None:

    with pytest.raises(CalledProcessError) as exc:
        generate_repo_with_merge_commit(
            tmp_path, with_conflict=True, secret_location=SecretLocation.CONFLICT_FILE
        )

    # AND the error message contains the Gitlab Token
    stderr = exc.value.stderr.decode()
    assert "GitLab Token" in stderr



from typing import Any, Dict, Optional, Union

import click

from .constants import IncidentStatus


LIGHT_GREY = (146, 146, 146)
ORANGE = (255, 128, 80)

STYLE: Dict[str, Dict[str, Any]] = {
    "nb_secrets": {"fg": "bright_blue", "bold": True},
    "filename": {"fg": "bright_yellow", "bold": True},
    "commit_info": {"fg": "bright_yellow", "bold": False},
    "patch": {"fg": "white"},
    "key": {"fg": "bright_white", "bold": True},
    "secret": {"fg": "bright_red"},
    "error": {"fg": "red"},
    "no_secret": {"fg": "white"},
    "detector": {"fg": "bright_yellow", "bold": True},
    "policy": {"fg": "cyan", "bold": True},
    "detector_line_start": {"fg": "cyan"},
    "line_count": {"fg": "white", "dim": True},
    "line_count_secret": {"fg": "yellow"},
    "progress": {"fg": "bright_yellow", "bold": False},
    "warning": {"fg": "yellow"},
    "heading": {"fg": "green"},
    "incident_validity": {"fg": "bright_yellow", "bold": True},
    "secret_type": {"fg": "bright_yellow", "bold": True},
    "occurrence_count": {"fg": "bright_yellow", "bold": True},
    "ignore_sha": {"fg": "bright_yellow", "bold": True},
}


def format_text(text: str, style: Dict[str, Any]) -> str:
    """Return the formatted text with the given style."""
    return click.style(
        text, fg=style["fg"], bold=style.get("bold", False), dim=style.get("dim", False)
    )


def pluralize(name: str, nb: int, plural: Union[str, None] = None) -> str:
    # Note: 0 is plural in english
    if nb == 1:
        return name
    return plural or (name + "s")


def format_line_count(line_count: Union[int, None], padding: int) -> str:
    """Return the padded line count."""
    if line_count is None:
        return " " * padding

    return " " * max(0, padding - len(str(line_count))) + str(line_count)


_VALIDITY_TEXT_FOR_ID = {
    "unknown": "Unknown",
    # cannot_check is the old ID for secrets for which there are no checkers
    "cannot_check": "Cannot Check",
    "no_checker": "No Checker",
    "failed_to_check": "Failed to Check",
    "not_checked": "Not Checked",
    "invalid": "Invalid",
    "valid": "Valid",
}


def translate_validity(validity_id: Optional[str]) -> str:
    if validity_id is None:
        validity_id = "unknown"
    # If we don't have a text for the validity_id, return it as is. We assume the text
    # of the ID is more valuable than a generic "Unknown" string
    return _VALIDITY_TEXT_FOR_ID.get(validity_id, validity_id)


def clip_long_line(
    content: str,
    max_length: int,
    before: bool = False,
    after: bool = False,
    min_length: int = 10,
    is_patch: bool = False,
) -> str:
    """
    Add a "…" character before and/or after the given string
    if it exceeds a maximum length.
    """
    ellipsis = "…"

    if is_patch:
        # If we are clipping a patch, move the first character out of `content` so that
        # we can keep it visible. It's important because this character is the patch
        # symbol.
        prefix = content[:1]
        content = content[1:]
        max_length -= 1
    else:
        prefix = ""

    content_length = len(content)
    if content_length > max_length:
        if before and after and content_length > max_length + 1:
            content = (
                ellipsis
                + content[
                    (content_length - max(max_length, min_length)) // 2
                    + 1 : (content_length + max(max_length, min_length)) // 2
                    - 1
                ]
                + ellipsis
            )
        elif after:
            content = content[: max(max_length - 1, min_length)] + ellipsis
        elif before:
            content = ellipsis + content[min(-max_length + 1, -min_length) :]
    return prefix + content


def file_info(
    filename: str,
    incident_count: int,
    incident_status: IncidentStatus = IncidentStatus.DETECTED,
) -> str:
    """Return the formatted file info (number of incidents + filename)."""
    return "\n{} {}: {} {} {}\n".format(
        format_text(">", STYLE["detector_line_start"]),
        format_text(filename, STYLE["filename"]),
        incident_count,
        pluralize("incident", incident_count, "incidents"),
        incident_status.value,
    )



from .commit import Commit
from .file import File, create_files_from_paths
from .scan_context import ScanContext
from .scan_mode import ScanMode
from .scannable import DecodeError, Scannable, StringScannable


__all__ = [
    "create_files_from_paths",
    "Commit",
    "DecodeError",
    "File",
    "ScanContext",
    "ScanMode",
    "Scannable",
    "StringScannable",
]



import os
from pathlib import Path
from typing import Callable
from unittest import mock

import pytest
from pytest import MonkeyPatch

from ggshield.core.git_hooks.ci.previous_commit import get_previous_commit_from_ci_env
from ggshield.utils.git_shell import EMPTY_SHA, git, is_valid_git_commit_ref
from ggshield.utils.os import cd
from tests.repository import Repository


def _setup_github_ci_env_new_branch(
    monkeypatch: MonkeyPatch,
    branch_name: str,
    head_sha: str,
    before_sha: str = EMPTY_SHA,
):
    monkeypatch.setenv("GITHUB_ACTIONS", "1")
    monkeypatch.setenv("GITHUB_PUSH_BEFORE_SHA", before_sha)
    monkeypatch.setenv("GITHUB_BASE_REF", "")
    monkeypatch.setenv("GITHUB_SHA", head_sha)
    monkeypatch.setenv("GITHUB_EVENT_NAME", "push")
    monkeypatch.setenv("GITHUB_REF_TYPE", "branch")
    monkeypatch.setenv("GITHUB_REF_NAME", branch_name)


def _setup_gitlab_ci_env_new_branch(
    monkeypatch: MonkeyPatch,
    branch_name: str,
    head_sha: str,
    before_sha: str = EMPTY_SHA,
):
    monkeypatch.setenv("GITLAB_CI", "1")
    monkeypatch.setenv("CI_COMMIT_BEFORE_SHA", before_sha)
    monkeypatch.setenv("CI_COMMIT_BRANCH", branch_name)
    monkeypatch.setenv("CI_COMMIT_SHA", head_sha)


def _setup_jenkins_ci_env_new_branch(
    monkeypatch: MonkeyPatch, branch_name: str, head_sha: str, before_sha: str = ""
):
    monkeypatch.setenv("JENKINS_HOME", "1")
    monkeypatch.setenv("GIT_PREVIOUS_COMMIT", before_sha)
    monkeypatch.setenv("GIT_BRANCH", branch_name)
    monkeypatch.setenv("GIT_COMMIT", head_sha)


def _setup_azure_ci_env_new_branch(
    monkeypatch: MonkeyPatch, branch_name: str, head_sha: str, before_sha: str = ""
):
    monkeypatch.setenv("BUILD_BUILDID", "1")
    monkeypatch.setenv("BUILD_SOURCEVERSION", before_sha)
    monkeypatch.setenv("BUILD_SOURCEBRANCHNAME", branch_name)


parametrized_ci_provider = pytest.mark.parametrize(
    "setup_ci_env",
    [
        _setup_github_ci_env_new_branch,
        _setup_gitlab_ci_env_new_branch,
        _setup_azure_ci_env_new_branch,
        _setup_jenkins_ci_env_new_branch,
    ],
    ids=["github", "gitlab", "azure_devops", "jenkins"],
)


@parametrized_ci_provider
@pytest.mark.parametrize("new_branch_commits_count", [0, 3])
def test_get_previous_commit_from_ci_env_new_branch(
    monkeypatch: MonkeyPatch,
    tmp_path: Path,
    setup_ci_env: Callable[[MonkeyPatch, str, str], None],
    new_branch_commits_count: int,
) -> None:
    """
    GIVEN   a repository with:
            - a commit pushed on the main branch
            - a new branch created from main, with or without commits
    WHEN    calculating the commit sha previous to the push,
            in a simulated CI env where new-branch is newly pushed
    THEN    the last commit sha of the parent branch is returned
    """
    # Setup main branch
    remote_repository = Repository.create(tmp_path / "remote", bare=True)
    repository = Repository.clone(remote_repository.path, tmp_path / "local")
    expected_sha = repository.create_commit()
    repository.push()

    # Setup new branch
    repository.create_branch("new-branch")
    head_sha = "HEAD"
    for _ in range(new_branch_commits_count):
        head_sha = repository.create_commit()

    with cd(repository.path), mock.patch.dict(os.environ, clear=True):
        # Simulate CI env
        setup_ci_env(monkeypatch, "new-branch", head_sha)
        found_sha = get_previous_commit_from_ci_env()

    assert found_sha is not None, "No previous commit SHA found"
    found_sha_evaluated = git(["rev-parse", found_sha], cwd=repository.path)
    assert found_sha_evaluated == expected_sha


@parametrized_ci_provider
def test_get_previous_commit_from_ci_env_new_repo(
    monkeypatch: MonkeyPatch,
    tmp_path: Path,
    setup_ci_env: Callable[[MonkeyPatch, str, str], None],
):
    """
    GIVEN   a new, empty repository
    WHEN    calculating the commit sha previous to the first push,
            in a simulated CI env
    THEN    None is returned
    """
    remote_repository = Repository.create(tmp_path / "remote", bare=True)
    repository = Repository.clone(remote_repository.path, tmp_path / "local")

    repository.create_commit()

    with cd(repository.path), mock.patch.dict(os.environ, clear=True):
        # Simulate CI env
        setup_ci_env(monkeypatch, "new-branch", "HEAD")
        assert get_previous_commit_from_ci_env() is None


@parametrized_ci_provider
def test_get_previous_commit_from_ci_env_sub_branch_forced_push(
    monkeypatch: MonkeyPatch,
    tmp_path: Path,
    setup_ci_env: Callable[[MonkeyPatch, str, str], None],
) -> None:
    """
    GIVEN   a repository with:
            - a commit pushed on the main branch
            - another branch created from main, with commits pushed
            - a local history rewritten
    WHEN    calculating the commit sha previous to the forced push,
            in a simulated CI env where new-branch is newly pushed
    THEN    the commit SHA (previous HEAD) returned does not exist anymore
            and a scan all is triggered instead
    """
    # Setup main branch
    remote_repository = Repository.create(tmp_path / "remote", bare=True)
    repository = Repository.clone(remote_repository.path, tmp_path / "local")
    repository.create_commit()
    repository.push()

    # Setup new branch
    branch_name = "new-branch"
    repository.create_branch(branch_name)
    before_sha = repository.create_commit()

    repository.git("commit", "--amend", "--allow-empty", "-m", "message")
    repository.remove_unreachable_commits()

    assert not is_valid_git_commit_ref(before_sha)
    assert before_sha != repository.get_top_sha()

    with cd(repository.path), mock.patch.dict(os.environ, clear=True):
        # Simulate CI env
        setup_ci_env(monkeypatch, branch_name, "HEAD", before_sha)
        found_sha = get_previous_commit_from_ci_env()

    assert found_sha is None


@parametrized_ci_provider
def test_get_previous_commit_from_ci_env_default_branch_forced_push(
    monkeypatch: MonkeyPatch,
    tmp_path: Path,
    setup_ci_env: Callable[[MonkeyPatch, str, str], None],
):
    """
    GIVEN   a new, empty repository
    WHEN    calculating the commit sha previous to the first push,
            in a simulated CI env
    THEN    None is returned
    """
    remote_repository = Repository.create(tmp_path / "remote", bare=True)
    repository = Repository.clone(remote_repository.path, tmp_path / "local")

    repository.create_commit()
    before_sha = repository.create_commit()

    repository.git("commit", "--amend", "--allow-empty", "-m", "message")
    repository.remove_unreachable_commits()

    assert not is_valid_git_commit_ref(before_sha)
    assert before_sha != repository.get_top_sha()

    with cd(repository.path), mock.patch.dict(os.environ, clear=True):
        # Simulate CI env
        setup_ci_env(monkeypatch, "new-branch", "HEAD", before_sha)
        assert get_previous_commit_from_ci_env() is None


@parametrized_ci_provider
def test_get_previous_commit_from_ci_env_all_commits_forced_push(
    monkeypatch: MonkeyPatch,
    tmp_path: Path,
    setup_ci_env: Callable[[MonkeyPatch, str, str], None],
):
    """
    GIVEN   a new, empty repository
    WHEN    calculating the commit sha previous to the first push,
            in a simulated CI env
    THEN    None is returned
    """
    remote_repository = Repository.create(tmp_path / "remote", bare=True)
    repository = Repository.clone(remote_repository.path, tmp_path / "local")

    before_sha = repository.create_commit()
    repository.git("commit", "--amend", "--allow-empty", "-m", "message")
    repository.remove_unreachable_commits()

    assert not is_valid_git_commit_ref(before_sha)
    assert before_sha != repository.get_top_sha()

    with cd(repository.path), mock.patch.dict(os.environ, clear=True):
        # Simulate CI env
        setup_ci_env(monkeypatch, "new-branch", "HEAD", before_sha)
        assert get_previous_commit_from_ci_env() is None



import json
from typing import Any, Dict, Iterable, List, cast

from pygitguardian.client import VERSIONS
from pygitguardian.models import SecretIncident

from ggshield import __version__ as ggshield_version
from ggshield.core.match_span import MatchSpan

from ..extended_match import ExtendedMatch
from ..secret_scan_collection import Result, Secret, SecretScanCollection
from .secret_output_handler import SecretOutputHandler


SCHEMA_URL = "https://docs.oasis-open.org/sarif/sarif/v2.1.0/errata01/os/schemas/sarif-schema-2.1.0.json"


class SecretSARIFOutputHandler(SecretOutputHandler):

    def _process_scan_impl(self, scan: SecretScanCollection) -> str:
        incident_details = (
            scan.get_incident_details(self.client)
            if self.with_incident_details and self.client
            else {}
        )
        dct = {
            "version": "2.1.0",
            "$schema": SCHEMA_URL,
            "runs": [
                {
                    "tool": {
                        "driver": {
                            "organization": "GitGuardian",
                            "name": "ggshield",
                            "informationUri": "https://github.com/GitGuardian/ggshield",
                            "version": ggshield_version,
                        },
                        "extensions": [
                            {
                                "name": "secret",
                                "version": VERSIONS.secrets_engine_version,
                            }
                        ],
                    },
                    "results": list(
                        _create_sarif_results(scan.get_all_results(), incident_details)
                    ),
                }
            ],
        }
        return json.dumps(dct)


def _create_sarif_results(
    results: Iterable[Result], incident_details: Dict[str, SecretIncident]
) -> Iterable[Dict[str, Any]]:
    """
    Creates SARIF result dicts for our Result instances. Creates one SARIF result dict
    per policy break.
    """
    for result in results:
        for secret in result.secrets:
            yield _create_sarif_result_dict(result.url, secret, incident_details)


def _create_sarif_result_dict(
    url: str,
    secret: Secret,
    incident_details: Dict[str, SecretIncident],
) -> Dict[str, Any]:
    # Prepare message with links to the related location for each match
    matches_str = ", ".join(
        f"[{m.match_type}]({id})" for id, m in enumerate(secret.matches)
    )
    matches_li = "\n".join(
        f"- [{m.match_type}]({id})" for id, m in enumerate(secret.matches)
    )
    extended_matches = cast(List[ExtendedMatch], secret.matches)
    message = f"Secret detected: {secret.detector}.\nMatches: {matches_str}"
    markdown_message = f"Secret detected: {secret.detector}\nMatches:\n{matches_li}"

    # Create dict
    dct = {
        "ruleId": secret.detector,
        "level": "error",
        "message": {
            "text": message,
            "markdown": markdown_message,
        },
        "locations": [
            _create_location_dict(url, [m.span for m in extended_matches]),
        ],
        "relatedLocations": [
            _create_related_location_dict(url, id, m)
            for id, m in enumerate(extended_matches)
        ],
        "partialFingerprints": {
            "secret/v1": secret.get_ignore_sha(),
        },
    }
    if secret.incident_url:
        dct["hostedViewerUri"] = secret.incident_url
        details = incident_details.get(secret.incident_url)
        if details is not None:
            dct["properties"] = {"incidentDetails": details.to_dict()}
    return dct


def _create_location_dict(
    url: str,
    match_spans: List[MatchSpan],
) -> Dict[str, Any]:
    # Create a span from the start of the first match to the end of the last match
    start_pos = min((x.line_index_start, x.column_index_start) for x in match_spans)
    end_pos = max((x.line_index_end, x.column_index_end) for x in match_spans)
    span = MatchSpan(
        line_index_start=start_pos[0],
        line_index_end=end_pos[0],
        column_index_start=start_pos[1],
        column_index_end=end_pos[1],
    )

    return {"physicalLocation": _create_physical_location_dict(url, span)}


def _create_related_location_dict(
    url: str,
    id: int,
    match: ExtendedMatch,
) -> Dict[str, Any]:
    return {
        "id": id,
        "physicalLocation": _create_physical_location_dict(url, match.span),
        "message": {"text": match.match_type},
    }


def _create_physical_location_dict(url: str, match_span: MatchSpan) -> Dict[str, Any]:
    return {
        "artifactLocation": {
            "uri": url,
        },
        "region": {
            "startLine": match_span.line_index_start + 1,
            "startColumn": match_span.column_index_start + 1,
            "endLine": match_span.line_index_end + 1,
            "endColumn": match_span.column_index_end + 1,
        },
    }



from abc import ABC, abstractmethod
from typing import Any, Sequence

from typing_extensions import Self

from ggshield.core.scan import Scannable


class ScannerUI(ABC):
    """
    An abstract class used by scanning code to notify user about progress or events
    during a scan
    """

    @abstractmethod
    def on_scanned(self, scannables: Sequence[Scannable]) -> None:
        raise NotImplementedError

    @abstractmethod
    def on_skipped(self, scannable: Scannable, reason: str) -> None:
        """
        Called when a scannable was skipped, `reason` explains why. If `reason` is empty
        then the user should not be notified of the skipped scannable (this happens for
        example when skipping empty files)
        """
        raise NotImplementedError

    @abstractmethod
    def __enter__(self) -> Self:
        raise NotImplementedError

    @abstractmethod
    def __exit__(self, *args: Any) -> None:
        raise NotImplementedError



from dataclasses import dataclass
from typing import List, Tuple


@dataclass
class VaultMount:
    """
    Model to represent a Vault mount.
    """

    name: str


@dataclass
class VaultKvMount(VaultMount):
    """
    Model to represent a Vault KV mount.
    """

    version: str


@dataclass
class VaultSecrets:
    """
    Model to hold secrets fetched from the vault.

    The aim is to include the secrets themselves and the not_fetched_paths
    list to communicate the paths that could not be fetched
    (permission denied, errors etc.).
    """

    secrets: List[Tuple[str, str]]
    not_fetched_paths: List[str]



from pygitguardian.models import Match

from ggshield.core.config.user_config import SecretConfig
from ggshield.verticals.secret import Results, SecretScanCollection
from ggshield.verticals.secret.output.secret_gitlab_webui_output_handler import (
    SecretGitLabWebUIOutputHandler,
    format_secret,
)
from tests.factories import SecretFactory


def test_format_secret():
    secret = SecretFactory(
        matches=[
            Match("AZERTYUIOP", "client_id", line_start=123),
            Match("abcdefghijk", "client_secret", line_start=456),
        ],
    )
    out = format_secret(secret)

    assert secret.detector in out
    assert "Validity: Valid" in out
    for match in secret.matches:
        assert match.match_type in out
        # match value itself must be obfuscated
        assert match.match not in out


def test_gitlab_web_ui_output_no_secrets():
    """
    GIVEN a content with no secret
    WHEN GitLabWebUIOutputHandler manipulates the corresponding scan
    THEN the error message is empty as expected and the status code is zero
    """
    output_handler = SecretGitLabWebUIOutputHandler(
        secret_config=SecretConfig(), verbose=False
    )
    scan = SecretScanCollection(
        id="scan",
        type="test",
        results=Results(results=[], errors=[]),
    )
    # call output handler
    exit_code = output_handler._get_exit_code(
        SecretScanCollection(
            id="outer_scan",
            type="outer_scan",
            results=Results(results=[], errors=[]),
            scans=[scan],
        )
    )
    error_msg = output_handler._process_scan_impl(scan=scan)

    assert exit_code == 0
    assert error_msg == ""



import shutil
import stat
import tarfile
from functools import lru_cache
from pathlib import Path
from typing import List, Optional, Tuple
from zipfile import ZipFile, ZipInfo


class UnsafeArchive(Exception):
    """
    Raised when an archive is considered unsafe.

    An archive is considered unsafe if it contains paths outside the extract directory
    or symbolic links pointing outside the extract directory.
    """

    def __init__(self):
        self.bad_paths: List[Path] = []
        self.bad_links: List[Tuple[Path, Path]] = []

    def __str__(self) -> str:
        lines: List[str] = ["Archive contains unsafe files."]
        if self.bad_paths:
            lines.append("Paths outside the archive root:")
            lines.extend(f"- {p}" for p in self.bad_paths)
        if self.bad_links:
            lines.append("Links pointing outside the archive root:")
            lines.extend(f"- {src} -> {target}" for src, target in self.bad_links)
        return "\n".join(lines)

    def add_bad_path(self, path: Path) -> None:
        self.bad_paths.append(path)

    def add_bad_link(self, path: Path, target: Path) -> None:
        self.bad_links.append((path, target))

    def has_fails(self) -> bool:
        return bool(self.bad_paths) or bool(self.bad_links)


def safe_unpack(archive: Path, extract_dir: Path) -> None:
    """
    If `archive` is safe, extract it in `extract_dir`. Otherwise, raise `UnsafeArchive`.
    """
    check_archive_content(archive)

    # unpack_archive does not know .whl files are zip files
    archive_format = "zip" if archive.suffix in {".whl", ".jar"} else None

    shutil.unpack_archive(archive, extract_dir, format=archive_format)


def check_archive_content(archive: Path) -> None:
    """
    Check `archive` safety, raise `UnsafeArchive` if it is unsafe.
    """
    if archive.suffix in {".zip", ".whl", ".jar"}:
        _check_zip_content(archive)
    else:
        _check_tar_content(archive)


@lru_cache
def _archive_root():
    """
    Returns a fake archive root, used by _is_bad_path().
    Can be whatever we want except the root directory, otherwise we won't detect a path
    outside the archive root.
    """

    # The call to `resolve()` is required on Windows to include the drive letter.
    return Path("/some/dir").resolve()


def _is_bad_path(path: Path) -> bool:
    """
    Verify `path` is not outside the archive root.

    Good paths: `foo`, `foo/../bar`
    Bad paths: `../foo`, `foo/../../bar`, `/foo/bar`
    """

    root = _archive_root()

    # Code can be simplified to use Path.is_relative_to() when we move to Python 3.9
    try:
        root.joinpath(path).resolve().relative_to(root)
        return False
    except ValueError:
        return True


def _is_bad_link(link: Path, target: Path) -> bool:
    """
    Given a link pointing to a target, verify it does not point outside the archive
    root.

    Good link example:
        link="foo/link", target="../file"
        Points to "file" in the archive root.

    Bad link example:
        link="foo/link", target="../../file"
        Points to "file" in the directory above the archive root.

    Another bad link example:
        link="foo/link", target="/etc/passwd"
        Points to "/etc/passwd"
    """

    # Links are interpreted relative to the directory containing them, so we join
    # target to the link parent
    target = link.parent.joinpath(target)

    return _is_bad_path(target)


def _check_tar_content(archive: Path) -> None:
    # This exception is populated with bad paths or bad links if we find some. It's only
    # raised if we found some.
    unsafe_archive = UnsafeArchive()

    with tarfile.open(archive) as tf:
        for info in tf:
            path = Path(info.name)
            if _is_bad_path(path):
                unsafe_archive.add_bad_path(path)
            elif info.issym() or info.islnk():
                target = Path(info.linkname)
                if _is_bad_link(path, target):
                    unsafe_archive.add_bad_link(path, target)

    if unsafe_archive.has_fails():
        raise unsafe_archive


def _zip_get_symlink_target(zip: ZipFile, info: ZipInfo) -> Optional[Path]:
    """
    If `info` points to a symbolic link inside `zip`, then return the link target.
    If it's a regular file, return None.
    """
    if not stat.S_ISLNK(info.external_attr >> 16):
        return None
    target = zip.open(info).read().decode()
    return Path(target)


def _check_zip_content(archive: Path) -> None:
    # This exception is populated with bad paths or bad links if we find some. It's only
    # raised if we found some.
    unsafe_archive = UnsafeArchive()

    with ZipFile(archive) as zip:
        for info in zip.infolist():
            path = Path(info.filename)
            if target := _zip_get_symlink_target(zip, info):
                if _is_bad_link(path, target):
                    unsafe_archive.add_bad_link(path, target)
            elif _is_bad_path(path):
                unsafe_archive.add_bad_path(path)

    if unsafe_archive.has_fails():
        raise unsafe_archive



# snapshottest: v1 - https://goo.gl/zC4yUc

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_handle_scan_error[single file exception] 1'] = '''
Error scanning. Results may be incomplete.
Add the following files to your paths-ignore:
- /home/user/too/long/file/name: filename:: [ErrorDetail(string='Ensure this field has no more than 256 characters.', code='max_length')]
'''

snapshots['test_handle_scan_error[too many documents] 1'] = '''
Error scanning. Results may be incomplete.
The following chunk is affected:
/example, /example, /example, /example, /example, /example, /example, /example, /example, /example, /example, /example, /example, /example, /example, /example, /example, /example, /example, /example, /example
400:Too many documents to scan
'''



import logging
from pathlib import Path
from typing import Any, List

import click

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
    create_output_handler,
)
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.cmd.utils.hooks import check_user_requested_skip
from ggshield.core import ui
from ggshield.core.client import create_client_from_config
from ggshield.core.git_hooks.prepush import collect_commits_refs
from ggshield.core.scan import ScanContext, ScanMode
from ggshield.utils.git_shell import (
    EMPTY_SHA,
    EMPTY_TREE,
    check_git_dir,
    get_list_commit_SHA,
)
from ggshield.verticals.secret.repo import scan_commit_range


logger = logging.getLogger(__name__)


@click.command()
@click.argument("prepush_args", nargs=-1, type=click.UNPROCESSED)
@add_secret_scan_common_options()
@click.pass_context
@exception_wrapper
def prepush_cmd(ctx: click.Context, prepush_args: List[str], **kwargs: Any) -> int:
    """
    Scan as a pre-push git hook all commits that are about to be pushed.
    """
    ctx_obj = ContextObj.get(ctx)
    ctx_obj.client = create_client_from_config(ctx_obj.config)
    config = ctx_obj.config

    if check_user_requested_skip():
        return 0

    local_commit, remote_commit = collect_commits_refs(prepush_args)
    logger.debug("refs=(%s, %s)", local_commit, remote_commit)

    if local_commit == EMPTY_SHA:
        ui.display_info("Deletion event or nothing to scan.")
        return 0

    if remote_commit == EMPTY_SHA:
        ui.display_info(
            f"New tree event. Scanning last {config.user_config.max_commits_for_hook} commits."
        )
        before = EMPTY_TREE
        after = local_commit
        cmd_range = f"{EMPTY_TREE} {local_commit}"
    else:
        before = remote_commit
        after = local_commit
        cmd_range = f"{remote_commit}...{local_commit}"

    commit_list = get_list_commit_SHA(
        cmd_range, max_count=config.user_config.max_commits_for_hook + 1
    )

    if not commit_list:
        ui.display_warning(
            "Unable to get commit range.\n"
            f"  before: {before}\n"
            f"  after: {after}\n"
            "Skipping pre-push hook\n",
        )
        return 0

    if len(commit_list) > config.user_config.max_commits_for_hook:
        ui.display_info(
            f"Too many commits. Scanning last {config.user_config.max_commits_for_hook} commits\n"
        )
        commit_list = commit_list[-config.user_config.max_commits_for_hook :]

    ui.display_verbose(f"Commits to scan: {len(commit_list)}")

    check_git_dir()

    scan_context = ScanContext(
        scan_mode=ScanMode.PRE_PUSH,
        command_path=ctx.command_path,
        target_path=Path.cwd(),
    )

    return_code = scan_commit_range(
        client=ctx_obj.client,
        cache=ctx_obj.cache,
        commit_list=commit_list,
        output_handler=create_output_handler(ctx),
        exclusion_regexes=ctx_obj.exclusion_regexes,
        secret_config=config.user_config.secret,
        scan_context=scan_context,
    )
    if return_code:
        ui.display_info(ctx_obj.client.remediation_messages.pre_push)
    return return_code



import re

from tests.functional.utils import run_ggshield


def test_cli_order():
    """
    Because it is more natural.
    """
    # GIVEN a call to ggshield hmsl help
    result = run_ggshield("hmsl", "-h", expected_code=0)
    # fingerprint, query and decrypt command follow each other in this order.
    assert re.search(
        "\n  fingerprint +[^\n]+\n  query +[^\n]+\n  decrypt", result.stdout
    ), result.stdout
